/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "muGUI.h"

 

void muGUIwidget_init ( muGUIwidget * fwid, unsigned char ftype, int fx, int fy, unsigned int fw, 
			unsigned int fh, char * fTitle ) 
{	
	
	/*Assign Value of the Widget*/		
	fwid->Type 	= ftype ;
	fwid->x 	= fx;
	fwid->y 	= fy;
	fwid->w 	= fw;
	fwid->h 	= fh;

	/*Copy the Widget Title*/	
	memset(fwid->Title,0,56);
	strcpy(fwid->Title,fTitle);

	/*On Create the widget need to be display*/
	fwid->NeedRefresh = TRUE;
	fwid->moved = TRUE;
	fwid->Status = MUGUI_WID_STATUS_SHOW;
	
	/*All the widget has the same external draw*/
	fwid->external_draw = muGUIwidget_external_draw;

	/*Test if Title bar is click*/
	fwid->isClick = muGUIwidget_isClick_TitleBar;
	
	/*Default*/
	fwid->image 		= NULL; 
	fwid->addImage 		= NULL;
	fwid->TextBuffer 	= NULL;
	
	
	switch ( ftype ) {
		
		/*If Widget Type is IMAGE*/
		case MUGUI_WID_IMAGE : {
			fwid->internal_draw = muGUIwidget_image_internal_draw;	
			fwid->addImage 		= muGUIwidget_attachImage;
			fwid->freeMem 		= muGUIwidget_FreeImage;
			break;	
		}

		/*If Widget Type is TEXT*/
		case MUGUI_WID_TEXT : {
		
			fwid->internal_draw = muGUIwidget_text_internal_draw;
			fwid->addText 		= muGUIwidget_addText;
			fwid->freeMem 		= muGUIwidget_FreeAllText;	
			
			/*create memory for the current text*/
			fwid->TextBuffer = ( muGUIChainedText * ) malloc ( sizeof ( struct _muGUIChainedText ) * sizeof(char) );

			/*Initialisation of the First Element of the Chained List*/
			fwid->TextBuffer->next = NULL;
			fwid->TextBuffer->previous = NULL;
		
			/*Compute the max number of lignes to display 
			22 is TitleBar Height
			11 is Font Height
			*/
			fwid->maxLines = (unsigned int)( (fh - 22 ) /  11 );
			
			/*At Start, display at line 0*/
			fwid->displayAtLine = 0 ;
			
			break;	
		}		
		
	
		
		/*If Widget Type is unknown*/
		default: {
			fwid->internal_draw = muGUIwidget_default_internal_draw;
			fwid->freeMem 		= muGUIwidget_FreeDefault;
		}
	}
	
		
	
}

/*
Create the outer box of the widget with the title Bar and text
and a button in the TitleBar
*/
void muGUIwidget_external_draw ( void * this , void * fgui ) {

	SDL_Rect rect;
	SDL_Color rgb;
	Uint32 color;
	
	muGUIwidget * zWid;		
	muGUI * zgui;
	
	zWid = (muGUIwidget*)this;		//Get the Widget handle
	zgui = (muGUI*)fgui;			//Get the GUI handle




	if ( zWid->Status == MUGUI_WID_STATUS_SHOW ) {
			zWid->h = zWid->previous_h;
		
	}


	/*Draw the Outer Rectangle*/
	rgb.r = 0x00;	//Set the color value
	rgb.g = 0x25;	
	rgb.b = 0x49;	
	color = SDL_MapRGB(zgui->MainScreen->format, rgb.r, rgb.g, rgb.b);	//Get the Color
	SDL_drawLine(zgui->MainScreen, zWid->x, 			zWid->y, 			zWid->x + zWid->w , zWid->y, 			color);
	SDL_drawLine(zgui->MainScreen, zWid->x + zWid->w,	zWid->y, 			zWid->x + zWid->w , zWid->y + zWid->h,	color);
	SDL_drawLine(zgui->MainScreen, zWid->x , 			zWid->y + zWid->h , zWid->x + zWid->w , zWid->y + zWid->h, 	color);
	SDL_drawLine(zgui->MainScreen, zWid->x , 			zWid->y, 			zWid->x , 			zWid->y + zWid->h, 	color);
	SDL_drawLine(zgui->MainScreen, zWid->x , zWid->y + 21, zWid->x + zWid->w - 2 , zWid->y + 21 , 	color);	
	
	/*Fill the Title Bar*/
	rgb.r = 0x00;	//Set the color values
	rgb.g = 0x65;	
	rgb.b = 0x89;	
	color = SDL_MapRGB(zgui->MainScreen->format, rgb.r, rgb.g, rgb.b);	//get the color
 
  	rect.x = zWid->x+1;	//Set the Rectangle
 	rect.y = zWid->y+1;
  	rect.w = zWid->w-2;
  	rect.h = 20;
  	SDL_FillRect(zgui->MainScreen, &rect, color);	//Draw rectangle

	/*Print the Title Bar*/
	PrintStringFont  (zgui->MainScreen, zgui->Font1,zWid->x + 2, zWid->y+2, zWid->Title );
	
	/*Draw the First Button*/
	rgb.r = 0xCC;	//Set the color values
	rgb.g = 0xCC;	
	rgb.b = 0xCC;	
	color = SDL_MapRGB(zgui->MainScreen->format, rgb.r, rgb.g, rgb.b);		//get the color
  	rect.x = zWid->x+zWid->w-22;	//set the rectangle
  	rect.y = zWid->y+3;
  	rect.w = 16;
  	rect.h = 16;
  	SDL_FillRect(zgui->MainScreen, &rect, color);	//Fill the rectangle
	
	
}

/*
Draw Nothing
*/
void muGUIwidget_null_external_draw ( void * this , void * fgui ) {

	muGUIwidget * zWid;		
	muGUI * zgui;
	
	zWid = (muGUIwidget*)this;		//Get the Widget handle
	zgui = (muGUI*)fgui;			//Get the GUI handle

	
	
	if ( zWid->Status == MUGUI_WID_STATUS_SHOW ) {
			zWid->h = zWid->previous_h;
		
	}
}



/*
Default Internal Widget Draw fonction : Fill Rectangle
*/
void muGUIwidget_default_internal_draw ( void * this , void * fgui ) {
	SDL_Rect rect;
	SDL_Color rgb;
	Uint32 color;

	muGUIwidget * zWid;
	muGUI * zgui;
	
	zWid = (muGUIwidget*)this;	//Get the Widget handle
	zgui = (muGUI*)fgui;		//Get the GUI handle
	
	/*Draw the default Widget background */
	rgb.r = 0xAA;
	rgb.g = 0xAA;	
	rgb.b = 0xAA;	
	color = SDL_MapRGB(zgui->MainScreen->format, rgb.r, rgb.g, rgb.b);
	rect.x = zWid->x+1;
	rect.y = zWid->y+22;
	rect.w = zWid->w-2;
	rect.h = zWid->h-22;
	SDL_FillRect(zgui->MainScreen, &rect, color);
  
	
	
}


/*
Default Free Widget
Do Nothing but Must Exist
*/
void muGUIwidget_FreeDefault ( void * this ) { }



/*
Test if a Mouse Clisk is inside the Title and if yes , where

*/
int  muGUIwidget_isClick_TitleBar ( void * this , void * fgui , int fx , int fy) {
	
	muGUIwidget * zWid;
	muGUI * zgui;
	
	zWid = (muGUIwidget*)this;	//Get the Widget handle
	zgui = (muGUI*)fgui;		//Get the GUI handle
	
	
	/*If the mouse click is inside the Title Bar*/
	if ( (( fx > zWid->x) && ( fx < ( zWid->x + zWid->w  ))) && 
		 (( fy > zWid->y) && ( fy < ( zWid->y + 22 )))	) {
		 	
		 	/*if the first button is Clicked */
		 	if ( fx > ( zWid->x + zWid->w - 20 )) return (MUGUI_WID_TITLEBAR_BUTTON_1) ;
		 	
		 	/*else Click is inside the Title*/
		 	else return ( MUGUI_WID_TITLEBAR_OTHER );
		 	
		}
	/* else the click is not inside the Title Bar*/	
	else return ( FALSE );	
		

}

/********************************************************************************************
								IMAGE WIDGET
*********************************************************************************************/

/*
Image widget Internal Draw : Blit to Main screen Surface
*/
void muGUIwidget_image_internal_draw ( void * this , void * fgui ) {

	SDL_Rect rect;

	muGUIwidget * zWid;
	muGUI * zgui;
	
	zWid = (muGUIwidget*)this;	//Get the Widget handle
	zgui = (muGUI*)fgui;		//Get the GUI handle
	
	/*	Define the rectangle to draw
		Only the interneal (inside part ) of the Widget */
	rect.x = zWid->x+1;
	rect.y = zWid->y+22;
	rect.w = zWid->w-2;
	rect.h = zWid->h-22;

	/*Copy to the GUI Surface*/		
	SDL_BlitSurface(zWid->image , NULL, zgui->MainScreen, &rect);

}


/*
Image widget Internal Draw FullScreen : Use the SDL_gfx
*/
void muGUIwidget_image_internal_draw_fullscreen ( void * this , void * fgui ) {


	muGUIwidget * zWid;
	muGUI * zgui;

	double	ratio_x;
	double  ratio_y;
	
		
	zWid = (muGUIwidget*)this;	//Get the Widget handle
	zgui = (muGUI*)fgui;		//Get the GUI handle
	
	
	
	ratio_x = (double)video_screen->w / (double)zWid->image->w;
	ratio_y = (double)video_screen->h / (double)zWid->image->h;

	SDL_FreeSurface ( zgui->MainScreen_FullScreen );
	zgui->MainScreen_FullScreen = zoomSurface(zWid->image, ratio_x, ratio_y, 0);
}



/*
Attach a previously created image to the Widget
*/
void muGUIwidget_attachImage ( void * this , SDL_Surface * surface ) {
	
	muGUIwidget * zWid;

	zWid = (muGUIwidget*)this;	//Get the Widget handle

	zWid->image = surface;		//Set the Widget Surface
	
}

/*
Free Memory for the IMAGE Widget
*/
void muGUIwidget_FreeImage ( void * this ) {
	
	muGUIwidget * zWid;
	
	zWid = (muGUIwidget*)this;								//Get the Widget handle
	
	SDL_FreeSurface ( zWid->image );						//Free the Surface
	printf("Widget %d : Free 1 SDL_Surface\n",zWid->ID);	//Debug Output
	
}



/********************************************************************************************
								TEXT WIDGET
*********************************************************************************************/

/*
Text widget Internal Draw : Print Text to the Widget
*/
void muGUIwidget_text_internal_draw ( void * this , void * fgui ) {
	
	SDL_Rect rect;
	SDL_Color rgb;
	Uint32 color;
	unsigned int boucle;
	unsigned int Ypos;
	unsigned int currentLine;

	muGUIwidget 		* zWid;
	muGUI 				* zgui;
	muGUIChainedText 	* currentText;
	
	zWid = (muGUIwidget*)this;		//Get the Widget handle
	zgui = (muGUI*)fgui;			//Get the GUI handle
	
	/*Draw the background*/	
	rgb.r = 0x00;
	rgb.g = 0x00;	
	rgb.b = 0x00;	
	color = SDL_MapRGB(zgui->MainScreen->format, rgb.r, rgb.g, rgb.b);
	rect.x = zWid->x+1;
	rect.y = zWid->y+22;
	rect.w = zWid->w-2;
	rect.h = zWid->h-22;
	SDL_FillRect(zgui->MainScreen, &rect, color);


	/* get the first text to be displayed */
	currentText = zWid->TextBuffer;		//Get the first Element of the List
	for ( boucle = 0 ; boucle < zWid->displayAtLine ; boucle++ ) currentText = currentText->next;
	
	/*initialisation*/
	Ypos = rect.y ;		
	currentLine = 0;

	/*Draw the text*/
	while ( ( currentText->next != NULL ) &&  ( currentLine < zWid->maxLines ) ) {
	
		PrintStringFont  (zgui->MainScreen, zgui->Font2,rect.x + 2, Ypos, currentText->text );
		Ypos = Ypos + FontHeight(zgui->Font2) ;

		currentText = currentText->next;
		currentLine = currentLine + 1;
	}
	
}

/*
Add a Text to the Widget : Append a text to the chained list
Step 1 - Get the number of lines of the Widget and the Last Element of the Chained List
Step 2 - Set the Last Element Values 
Step 3 - Create a new element at the End of the chained List
*/
void muGUIwidget_addText ( void * this , char * ftext ) {

	muGUIChainedText 	* currentText;
	muGUIChainedText 	* previousText;
	muGUIChainedText 	* nextText;		
	muGUIwidget 		* zWid;
	unsigned int 		nbLigne;
	
	zWid = (muGUIwidget*)this;		//Get the Widget handle

	/*initialisation*/
	currentText = zWid->TextBuffer;		//Get the first Element of the List
	previousText = NULL;				//so there is no previous text
	nbLigne = 1;						//
	

	
	/* Get the number of elements ( lines ) of the chained List and the last Element*/
	while ( currentText->next != NULL ) {	//While the next element exist
		previousText = currentText;			//the previous element became the current element
		currentText = currentText->next;	//the current element became the next element
		nbLigne = nbLigne + 1 ;				//Increment the line number
	}
	
	/*Memory allocation for the String*/
	currentText->text = (char * )malloc ( strlen ( ftext ) + 1 );
	if ( currentText->text == NULL )	return;
	
	/*Copy the String to the Chained List*/
	strcpy( currentText->text , ftext );
	
	/*Set the previous element of the current element */
	currentText->previous = previousText;

	/*Compute the First Element to be displayed */
	if ( nbLigne < zWid->maxLines ) zWid->displayAtLine = 0;
	else zWid->displayAtLine = nbLigne - zWid->maxLines ;

	/* the Widget need to be refresh */
	zWid->NeedRefresh = TRUE;



	/* Memory Allocation for the next element*/
	currentText->next =  malloc ( sizeof ( struct _muGUIChainedText ) * sizeof(char) );

	/*If a memory allocation occurs...*/
	if ( currentText->next == NULL ) {
		printf("GUI:Fatal while alloc Memory in Widget Text\n");		//Print a error Message
		currentText->next = NULL;										//Set the next element to NULL
		return;															//exit
	}
	/*else the memory allocation is correct*/
	else  {
		nextText = currentText->next;
		nextText->next = NULL;
		nextText->previous = currentText;
		
	}

}


/*
Free Memory for the TEXT Widget
*/
void muGUIwidget_FreeAllText ( void * this ) {

	muGUIChainedText 	* currentText;
	muGUIwidget 		* zWid;
	unsigned int 		nbLigne;
	
	zWid = (muGUIwidget*)this;			//Get the Widget handle
	currentText = zWid->TextBuffer;		//Get the first element of the chained list

	/* initialisation*/
	nbLigne = 0;
	
	/* While exist a next element */	
	while ( currentText->next != NULL ) {
		
		/* Free the String of the current element*/
		if ( currentText->text != NULL ) free ( currentText->text );
		
		/* And free the previous element of the List */
		if ( currentText->previous != NULL ) {
			free(currentText->previous);
			nbLigne++;
		}
		
		/* goto to next element*/
		currentText = currentText->next;
		
	}
	
	printf("Widget %d : Free %d lines \n",zWid->ID,nbLigne);		//Debug Output
	
}


/*
Debug Function
*/
void muGUIwidget_printAllText ( void * this ) {

	muGUIChainedText * currentText;
	muGUIwidget * zWid;

	
	zWid = (muGUIwidget*)this;
	currentText = zWid->TextBuffer;
	
	while ( currentText->next != NULL ) {
		printf("-%s-\n",currentText->text);
		currentText = currentText->next;
	}

}

